import os

import pandas as pd

import latdraw.lattice as lattice

try:
    import tfs

    TFS = True
except ImportError:
    TFS = False

try:
    import pand8

    MAD8 = True
except ImportError:
    MAD8 = False


COMMON_OPTICS_NAMES = {
    "s": ["s", "SUML", "z"],
    "beta_x": ["BETX", "beta_x"],
    "beta_y": ["BETY", "beta_y"],
    "alpha_x": ["ALFX", "alpha_x"],
    "alpha_y": ["ALFY", "alpha_y"],
    "dx": ["DX", "disp_x", "Dx"],
    "dy": ["DY", "disp_y", "Dy"],
}


class UnknownOpticsAccessor(RuntimeError):
    pass


def _get_optics_from_object(obj, key):
    for name in COMMON_OPTICS_NAMES[key]:
        try:
            return obj[name]
        except KeyError:
            continue
    raise UnknownOpticsAccessor(f"Unknown Optics key: {key}")


def coerce_optics(some_optics) -> tuple[list, list, list]:
    assert not isinstance(some_optics, lattice.Beamline)

    try:
        some_optics = os.fspath(some_optics)
    except TypeError:
        pass
    else:
        some_optics = _load_optics(some_optics)

    result = {}
    for key, keylist in COMMON_OPTICS_NAMES.items():
        for key2 in keylist:
            result[key] = _get_optics_from_object(some_optics, key)
    return pd.DataFrame.from_dict(result)


def _load_optics(some_optics):
    try:
        return tfs.read(some_optics)
    except tfs.errors.TfsFormatError:
        pass

    return pand8.read(some_optics)



"""Top-level package for latdraw."""

__author__ = """Stuart Derek Walker"""
__email__ = 'stuart.walker@desy.de'
__version__ = '0.1.0'


from latdraw.interfaces import read, read_bdsim_survey, read_mad8, read_madx
from latdraw.latdraw import draw, draw_survey
from latdraw.plot import subplots_with_lattice, subplots_with_lattices, plot_optics, two_axes_figure

from . import optics

# from latdraw.lattice import Beamline

import warnings

try:
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        import ocelot
    from latdraw.interfaces import lattice_from_ocelot
except ImportError:
    pass



import os
from typing import Union

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

import latdraw
from latdraw import lattice
import latdraw.interfaces as interfaces


def _get_lattice(lattice_or_path_to_one) -> lattice.Beamline:
    try:
        fpath = os.fspath(lattice_or_path_to_one)
    except TypeError:
        lattice = lattice_or_path_to_one
    else:
        return latdraw.read(fpath)

    try:
        lattice = interfaces.madx_twiss_to_beamline(lattice_or_path_to_one)
    except:
        pass

    return lattice


def subplots_with_lattice(
        lattice: Union[str, lattice.Beamline, pd.DataFrame], s_offset=0, nrows: int = 1, gridspec_kw=None, **kwargs
):
    # height_ratios = [1, 0.5, 0.5, 1]
    # the_gridspec_kw = {"height_ratios": height_ratios,
    #                    "hspace": 0.05}
    # if gridspec_kw is None:
    #     gridspec_kw = {}
    # the_gridspec_kw |= gridspec_kw

    # Plot goes at the top
    pattern = [lattice]
    pattern.extend(nrows * [None])
    return subplots_with_lattices(pattern, s_offset=s_offset, **kwargs)


def subplots_with_lattices(pattern, s_offset=0, **draw_kwargs) -> tuple[plt.Figure, list[plt.Axes]]:
    pattern = np.array(pattern, dtype=object)

    height_ratios = np.full_like(pattern, 1.0, dtype=float)
    # Get indices of where machines should be plotted
    indices = [index for (index, value) in enumerate(pattern) if value is not None]
    height_ratios[indices] = 0.25

    the_gridspec_kw = {"height_ratios": height_ratios, "hspace": 0.05}

    fig, axes = plt.subplots(nrows=len(pattern), sharex=True, gridspec_kw=the_gridspec_kw)

    for lattice, ax in zip(pattern, axes):
        if lattice is None:
            continue

        lattice = _get_lattice(lattice)
        latdraw.draw(fig, ax, lattice, s_offset=s_offset, **draw_kwargs)

        ax.set_yticks([], [])

        ax.tick_params(top=False, bottom=False, left=False, right=False, labelleft=False, labelbottom=False)

        ax.spines['left'].set_visible(False)
        ax.spines['right'].set_visible(False)

        # ax.spines['top'].set_visible(True)
        # ax.spines['bottom'].set_visible(False)

        ax.set_ylim(-0.25, 0.25)

    return fig, axes


def plot_optics(beamline, s, betax, betay):
    fig, (mx, axbx) = latdraw.subplots_with_lattice(beamline, nrows=1)

    axbx.plot(s, betax, label=r"$x$")
    axbx.plot(s, betay, label=r"$y$")

    axbx.legend()
    axbx.set_xlabel("$s$ / m")
    axbx.set_ylabel(r"$\beta_{x,y}$ / m")


def plot_optics(some_beamline, some_optics, title=""):
    bl = interfaces.coerce_beamline(some_beamline)
    optics = latdraw.optics.coerce_optics(some_optics)

    fig, axes = subplots_with_lattice(bl, nrows=2)

    fig, (mx, ax1, ax2) = fig, axes

    ax1.plot(optics.s, optics.beta_x, label=r"$\beta_x$")
    ax1.plot(optics.s, optics.beta_y, label=r"$\beta_y$")
    ax1.legend()
    beta_label(ax1)

    ax2.plot(optics.s, optics.dx, label=r"$D_x$")
    ax2.plot(optics.s, optics.dy, label=r"$D_y$")
    dispersion_label(ax2)
    s_label(ax2)
    ax2.legend()

    mx.set_title(title)

    return fig, axes


def simple_figure(some_beamline, title="", **drawkwargs) -> tuple[plt.Figure, tuple[plt.Axes, plt.Axes]]:
    bl = interfaces.coerce_beamline(some_beamline)

    fig, axes = subplots_with_lattice(bl, nrows=1, **drawkwargs)
    s_label(axes[-1])
    axes[0].set_title(title)
    # fig.suptitle(title)
    return fig, axes

def two_axes_figure(some_beamline, title="") -> tuple[plt.Figure, tuple[plt.Axes, plt.Axes, plt.Axes]]:
    bl = interfaces.coerce_beamline(some_beamline)

    fig, axes = subplots_with_lattice(bl, nrows=2)
    s_label(axes[-1])
    axes[0].set_title(title)
    # fig.suptitle(title)
    return fig, axes


def three_axes_figure(some_beamline, title="") -> tuple[plt.Figure, tuple[plt.Axes, plt.Axes, plt.Axes, plt.Axes]]:
    bl = interfaces.coerce_beamline(some_beamline)

    fig, axes = subplots_with_lattice(bl, nrows=3)
    s_label(axes[-1])
    axes[0].set_title(title)
    return fig, axes


def four_axes_figure(some_beamline, title="") -> tuple[plt.Figure, tuple[plt.Axes, plt.Axes, plt.Axes, plt.Axes]]:
    bl = interfaces.coerce_beamline(some_beamline)

    fig, axes = subplots_with_lattice(bl, nrows=4)
    s_label(axes[-1])
    axes[0].set_title(title)
    return fig, axes


def beta_label(ax, subscript=""):
    if not subscript:
        ax.set_ylabel(r"$\beta$ / m")
    else:
        ax.set_ylabel(rf"$\beta_{subscript}$ / m")


def alpha_label(ax, subscript=""):
    if not subscript:
        ax.set_ylabel(r"$\alpha$ / m")
    else:
        ax.set_ylabel(rf"$\alpha_{subscript}$")


def dispersion_label(ax, subscript=""):
    if not subscript:
        ax.set_ylabel(r"$D$ / m")
    else:
        ax.set_ylabel(rf"$D_{{{subscript}}}$ / m")        


def s_label(ax) -> None:
    ax.set_xlabel(r"$s$ / m")


def energy_label(ax, unit="GeV"):
    ax.set_ylabel(f"$E$ / {unit}")


def draw_vline(axes: list[plt.Axes], x: float, **vline_kwargs) -> None:
    for ax in axes:
        ax.axvline(x, **vline_kwargs)


def compare_two_optics(some_beamline1, optics1, some_beamline2, optics2, label1="", label2="", title="", dimension="x"):
    bl1 = interfaces.coerce_beamline(some_beamline1)
    optics1 = latdraw.optics.coerce_optics(optics1)

    bl2 = interfaces.coerce_beamline(some_beamline2)
    optics2 = latdraw.optics.coerce_optics(optics2)

    fig, axes = subplots_with_lattice(bl1, nrows=2)
    fig, (mx, ax1, ax2) = fig, axes

    (l1,) = ax1.plot(optics1.s, optics1[f"beta_{dimension}"], label=label1)
    beta_label(ax1)

    ax1.plot(optics2.s, optics2[f"beta_{dimension}"], label=label2)  # color=l1.get_color(),
    # ax1.plot(optics2.s, optics2.beta_y)
    if label1 or label2:
        ax1.legend()

    (l1,) = ax2.plot(optics1.s, optics1[f"alpha_{dimension}"])
    (l1,) = ax2.plot(optics2.s, optics2[f"alpha_{dimension}"])

    ax2.set_ylabel(rf"$\alpha_{dimension}$")

    beta_label(ax1, dimension)
    s_label(ax2)

    return fig, axes

    # ax2.plot(optics.s, optics.dx, label=r"$D_x$")
    # ax2.plot(optics.s, optics.dy, label=r"$D_y$")
    # dispersion_label(ax2)
    # s_label(ax2)
    # ax2.legend()



import os
from pathlib import Path
from typing import Mapping, Sequence, Tuple, Callable, Any, Generator, Iterator

import pand8
import pandas as pd
import tfs

try:
    from ocelot.cpbd import elements as ole
except ImportError:
    pass


import latdraw.lattice as lattice


try:
    import ocelot
    OCELOT = True
except ImportError:
    OCELOT = False


def flatten(iterable: Iterator[Any]) -> Generator[Any, None, None]:
    """Flatten arbitrarily nested iterable.
    Special case for strings that avoids infinite recursion.  Non iterables passed
    as arguments are yielded.

    :param iterable: Any iterable to be flattened.
    :raises: RecursionError

    """

    def _flatten(iterable):
        try:
            for item in iterable:
                try:
                    iter(item)
                except TypeError:
                    yield item
                else:
                    yield from _flatten(item)
        except TypeError:  # If iterable isn't actually iterable, then yield it.
            yield iterable

    try:
        yield from _flatten(iterable)
    except RecursionError:
        raise RecursionError("Maximum recusion reached.  Possibly trying" " to flatten an infinitely nested iterable.")


class FileTypeError(RuntimeError):
    pass


class UnknownBeamlineDescription(RuntimeError):
    pass


class UnknownElementType(RuntimeError):
    pass


def read(fname):
    """Generatl reader function"""
    try:
        return read_madx(fname)
    except tfs.errors.TfsFormatError:
        pass

    return read_mad8(fname)


def read_madx(fname):
    fname = os.fspath(fname)  # Accept any pathlike object

    df = tfs.read(fname)

    file_type = df.headers["TYPE"]
    if file_type not in {"SURVEY", "TWISS"}:
        raise FileTypeError(f"Unsupported TFS TYPE in header: {file_type}")

    latdraw_sequence = _loop_madx_df(df, survey=file_type == "SURVEY")
    return lattice.Beamline(latdraw_sequence)


def read_madx_survey(survey, twiss=None):
    # Optional twiss so for example the quads etc. point in the correct direction...
    pass


def madx_twiss_to_beamline(twiss):
    return lattice.Beamline(_loop_madx_df(twiss, survey=False))


def madx_survey_to_beamline(survey):
    return lattice.Beamline(_loop_madx_df(survey, survey=True))


def _loop_madx_df(tfs_df, survey):
    for tup in tfs_df.itertuples():
        name = tup.NAME
        length = tup.L

        z = tup.Z if survey else tup.S
        position = tup.X, tup.Y, z
        keyword = tup.KEYWORD

        if not survey and keyword.endswith("POLE"):
            k1 = tup.K1L / length
            k2 = tup.K2L / length
            k3 = tup.K3L / length
        else:
            k1 = k2 = k3 = 0

        if keyword == "DRIFT":
            yield lattice.Drift(name, length)
        elif keyword == "RBEND":
            yield lattice.RBend(name, length, tup.ANGLE)
        elif keyword == "SBEND":
            yield lattice.SBend(name, length, tup.ANGLE)
        elif keyword == "QUADRUPOLE":
            yield lattice.Quadrupole(name, length, k1)
        elif keyword == "SEXTUPOLE":
            yield lattice.Sextupole(name, length, k2)
        elif keyword == "OCTUPOLE":
            yield lattice.Sextupole(name, length, k3)
        elif keyword == "HKICKER":
            yield lattice.HKicker(name, length, tup.ANGLE)
        elif keyword == "VKICKER":
            yield lattice.VKicker(name, length, tup.ANGLE)
        elif keyword == "KICKER":
            yield lattice.Kicker(name, length, tup.ANGLE)
        elif keyword == "MARKER":
            yield lattice.Marker(name)
        elif keyword == "MONITOR":
            yield lattice.Monitor(name)
        else:
            raise UnknownElementType(f"NAME={name}, KEYWORD={keyword}")


def read_mad8(fname):
    fname = os.fspath(fname)  # Accept any pathlike object

    df = pand8.read(fname)

    file_type = df.attrs["DATAVRSN"]
    if file_type == "SURVEY":
        survey = True
    elif file_type == "TWISS":
        survey = False
    else:
        raise TFSTypeError(f"Unsupported MAD8 File DATAVRSN in header: {file_type}")

    latdraw_sequence = list(_loop_mad8_df(df, survey))
    return lattice.Beamline(latdraw_sequence)


def _loop_mad8_df(mad8_df, is_survey):
    for tup in mad8_df.itertuples():
        name = tup.NAME
        length = tup.L

        z = tup.Z if is_survey else tup.SUML
        position = tup.X, tup.Y, z
        keyword = tup.KEYWORD

        if keyword == "":  # Skip this blank element that starts every mad8 lattice?
            pass
        elif keyword == "DRIF":
            yield lattice.Drift(name, length)
        elif keyword == "RBEN":
            yield lattice.RBend(name, length, tup.ANGLE)
        elif keyword == "SBEN":
            yield lattice.SBend(name, length, tup.ANGLE)
        elif keyword == "QUAD":
            yield lattice.Quadrupole(name, length, tup.K1)
        elif keyword == "SEXT":
            yield lattice.Sextupole(name, length, tup.K2)
        elif keyword == "OCTU":
            yield lattice.Sextupole(name, length, tup.K3)
        elif keyword == "HKIC":
            yield lattice.HKicker(name, length, tup.ANGLE)
        elif keyword == "VKIC":
            yield lattice.VKicker(name, length, tup.ANGLE)
        elif keyword == "KICK":
            yield lattice.Kicker(name, length, tup.ANGLE)
        elif keyword == "MARK":
            yield lattice.Marker(name)
        elif keyword == "MONI":
            yield lattice.Monitor(name)
        elif keyword == "SOLE":
            yield lattice.Solenoid(name, length, 0.0)
        elif keyword == "ECOL":
            yield lattice.Collimator(name, length)
        elif keyword == "LCAV":
            yield lattice.RFCavity(name, length)
        elif keyword == "MATR":
            yield lattice.GenericMap(name, length)
        else:
            raise UnknownElementType(f"NAME={name}, KEYWORD={keyword}")


def read_bdsim_survey(fname, straighten=False):
    bdsim_survey_df = pd.read_csv(
        fname,
        skiprows=1,
        delim_whitespace=True,
        skipfooter=2,
        # engine="python"  # Because of use of skipfooter
    )
    new_columns = []
    for column_name in bdsim_survey_df.columns:
        name_without_units, *_ = column_name.split("[")
        new_columns.append(name_without_units)
    bdsim_survey_df.columns = new_columns

    latdraw_sequence = list(_loop_bdsim_survey_df(bdsim_survey_df, straighten))
    return lattice.Beamline(latdraw_sequence)


def _loop_bdsim_survey_df(bdsim_survey_df, straighten=False):
    ignoreable_types = {"dipolefringe"}
    for tup in bdsim_survey_df.itertuples():
        name = tup.Name
        length = tup.ChordLength

        if straighten:
            position = 0, 0, tup.SEnd
        else:
            position = tup.X, tup.Y, tup.Z

        keyword = tup.Type

        if keyword in ignoreable_types:
            continue

        if keyword == "drift":
            yield lattice.Drift(name, length)
        elif keyword == "rbend":
            yield lattice.RBend(name, length, tup.Angle)
        elif keyword == "sbend":
            yield lattice.SBend(name, length, tup.Angle)
        elif keyword == "quadrupole":
            yield lattice.Quadrupole(name, length, tup.k1)
        elif keyword == "sextupole":
            yield lattice.Sextupole(name, length, tup.k2)
        elif keyword == "octupole":
            yield lattice.Sextupole(name, length, tup.k3)
        elif keyword == "hkicker":
            yield lattice.HKicker(name, length, tup.Angle)
        elif keyword == "vkicker":
            yield lattice.VKicker(name, length, tup.Angle)
        elif keyword == "kicker":
            yield lattice.Kicker(name, length, tup.Angle)
        else:
            raise UnknownElementType(f"NAME={name}, KEYWORD={keyword}")


def lattice_from_ocelot(ocelot_lattice):
    seq = ocelot_lattice

    try:
        seq = ocelot_lattice.sequence
    except AttributeError:
        pass

    beamline = lattice.Beamline(_loop_lattice_from_ocelot(seq))
    return beamline


def coerce_beamline(some_beamline):
    try:
        some_beamline = Path(some_beamline)
    except TypeError:
        pass
    else:
        return read(some_beamline)

    if OCELOT:
        return lattice_from_ocelot(some_beamline)

    raise UnknownBeamlineDescription


def _loop_lattice_from_ocelot(ocelot_sequence):
    assert not isinstance(ocelot_sequence, str)

    for ele in flatten(ocelot_sequence):
        name = ele.id
        l = ele.l
        if isinstance(ele, ole.Marker):
            yield lattice.Marker(name)
        elif isinstance(ele, ole.Monitor):
            yield lattice.Monitor(name)
        elif isinstance(ele, ole.Drift):
            yield lattice.Drift(name, l)
        elif isinstance(ele, ole.RBend):
            yield lattice.RBend(name, l, ele.angle)
        elif isinstance(ele, ole.SBend):
            yield lattice.SBend(name, l, ele.angle)
        elif isinstance(ele, ole.Quadrupole):
            yield lattice.Quadrupole(name, l, ele.k1)
        elif isinstance(ele, ole.Sextupole):
            yield lattice.Sextupole(name, l, ele.k2)
        elif isinstance(ele, ole.TDCavity):
            yield lattice.TransverseDeflectingCavity(name, l, ele.v)
        elif isinstance(ele, ole.Vcor):
            yield lattice.VKicker(name, l, ele.angle)
        elif isinstance(ele, ole.Hcor):
            yield lattice.HKicker(name, l, ele.angle)
        elif isinstance(ele, ole.Cavity):
            yield lattice.RFCavity(name, l, ele.v, ele.phi)
        elif isinstance(ele, ole.Solenoid):
            yield lattice.Solenoid(name, l, ele.k)
        elif isinstance(ele, ole.Undulator):
            undulator_length = ele.nperiods * ele.lperiod
            yield lattice.Undulator(name, undulator_length, kx=ele.Kx, ky=ele.Ky)
        else:
            raise UnknownElementType(name, type(ele))



import numpy as np
import pytest

from latdraw import lattice


POSITION = [1, 2, 3]
LENGTH = 0.5

def test_init_Drift():
    name = "drift"
    drift = lattice.Drift(name, POSITION, LENGTH)

    assert drift.name == name
    assert drift.position == POSITION
    assert drift.length == LENGTH

def test_init_SBend():
    name = "sbend"
    angle = np.pi / 10
    sbend = lattice.SBend(name, POSITION, LENGTH, angle)

    assert sbend.name == name
    assert sbend.position == POSITION
    assert sbend.length == LENGTH
    assert sbend.angle == angle

def test_init_RBend():
    name = "rbend"
    angle = np.pi / 10
    sbend = lattice.RBend(name, POSITION, LENGTH, angle)

    assert sbend.name == name
    assert sbend.position == POSITION
    assert sbend.length == LENGTH
    assert sbend.angle == angle

def test_init_Quadrupole():
    name = "quadrupole"
    k1 = 0.005
    sbend = lattice.Quadrupole(name, POSITION, LENGTH, k1)

    assert sbend.name == name
    assert sbend.position == POSITION
    assert sbend.length == LENGTH
    assert sbend.k1 == k1

def test_init_Sextupole():
    name = "sextupole"
    k2 = 0.005
    sbend = lattice.Sextupole(name, POSITION, LENGTH, k2)

    assert sbend.name == name
    assert sbend.position == POSITION
    assert sbend.length == LENGTH
    assert sbend.k2 == k2

def test_init_Octupole():
    name = "octupole"
    k3 = 0.005
    sbend = lattice.Octupole(name, POSITION, LENGTH, k3)

    assert sbend.name == name
    assert sbend.position == POSITION
    assert sbend.length == LENGTH
    assert sbend.k3 == k3
    
def test_init_VKicker():
    name = "vkicker"
    angle = 0.005
    sbend = lattice.VKicker(name, POSITION, LENGTH, angle)

    assert sbend.name == name
    assert sbend.position == POSITION
    assert sbend.length == LENGTH
    assert sbend.angle == angle

def test_init_VKicker():
    name = "vkicker"
    angle = 0.005
    sbend = lattice.HKicker(name, POSITION, LENGTH, angle)

    assert sbend.name == name
    assert sbend.position == POSITION
    assert sbend.length == LENGTH
    assert sbend.angle == angle
    



"""Main module."""


from math import copysign

import numpy as np
import matplotlib.patches as patches
import matplotlib.pyplot as plt
import matplotlib as mpl
from adjustText import adjust_text


import latdraw.lattice as lattice

MAGNET_WIDTH = 0.1

DEFAULT_COLOUR_MAP = {
    "Quadrupole": "red",
    "SBend": "blue",
    "RBend": "blue",
    "RFCavity": "orange",
    "Marker": None,
    "Drift": None,
    "Monitor": None,
    "HKicker": "purple",
    "VKicker": "purple",
    "Kicker": "purple",
    "Cavity": "orange",
    "Collimator": "black",
    "GenericMap": "gray",
    "Solenoid": "g",
    "Sextupole": "g",
    "TransverseDeflectingCavity": "magenta",
    "Undulator": "cyan",
}


# dict_plot = {Quadrupole: {"scale": 0.7, "color": "r", "edgecolor": "r", "label": "quad"},
#              Sextupole: {"scale": 0.5, "color": "g", "edgecolor": "g", "label": "sext"},
#              Octupole: {"scale": 0.5, "color": "g", "edgecolor": "g", "label": "oct"},
#              Cavity: {"scale": 0.7, "color": "orange", "edgecolor": "lightgreen", "label": "cav"},
#              TWCavity: {"scale": 0.7, "color": "orange", "edgecolor": "lightgreen", "label": "twcav"},
#              Bend: {"scale": 0.7, "color": "lightskyblue", "edgecolor": "k", "label": "bend"},
#              RBend: {"scale": 0.7, "color": "lightskyblue", "edgecolor": "k", "label": "bend"},
#              SBend: {"scale": 0.7, "color": "lightskyblue", "edgecolor": "k", "label": "bend"},
#              Matrix: {"scale": 0.7, "color": "pink", "edgecolor": "k", "label": "mat"},
#              Multipole: {"scale": 0.7, "color": "g", "edgecolor": "k", "label": "mult"},
#              Undulator: {"scale": 0.7, "color": "pink", "edgecolor": "k", "label": "und"},
#              Monitor: {"scale": 0.5, "color": "orange", "edgecolor": "orange", "label": "mon"},
#              Hcor: {"scale": 0.7, "color": "c", "edgecolor": "c", "label": "cor"},
#              Vcor: {"scale": 0.7, "color": "c", "edgecolor": "c", "label": "cor"},
#              Drift: {"scale": 0., "color": "k", "edgecolor": "k", "label": ""},
#              Marker: {"scale": 0., "color": "k", "edgecolor": "k", "label": "mark"},
#              Solenoid: {"scale": 0.7, "color": "g", "edgecolor": "g", "label": "sol"},
#              TDCavity: {"scale": 0.7, "color": "magenta", "edgecolor": "g", "label": "tds"},
#              UnknownElement: {"scale": 0.7, "color": "g", "edgecolor": "g", "label": "unk"},
#              XYQuadrupole: {"scale": 0.7, "color": "r", "edgecolor": "r", "label": "xyquad"},
#              Aperture: {"scale": 0.7, "color": "g", "edgecolor": "g", "label": "ap"},
#              }


def draw_survey(
    fig, axes, sequence, colour_map=None, annotate=True, dimension="x", magnet_width=MAGNET_WIDTH, **drawlinekw
):
    try:
        survey = sequence.survey()
    except AttributeError:
        survey = sequence

    draw_survey_line(axes, survey["z"], survey[dimension], **drawlinekw)

    if colour_map is None:
        colour_map = DEFAULT_COLOUR_MAP

    patch_ids_to_annotations = {}
    patch_ids_to_elements = {}

    # import ipdb; ipdb.set_trace()

    for row in survey.itertuples():
        colour = DEFAULT_COLOUR_MAP[row.keyword]
        alpha = 1

        if not row.active:
            alpha = 0.25

        if colour is None:
            alpha = 0.0

        x, y, z = row.x, row.y, row.z
        # This is the end basically.
        if dimension == "y":
            x = y
        elif dimension != "x":
            raise ValueError(f"Unrecognised dimension type {dimension}.")
            
            
        length = row.length
        patch_start_x = z - length
        patch_width = length
        patch_start_y = x - 2 * magnet_width
        patch_height = 4 * magnet_width

        try:
            patch_height = copysign(patch_height * 0.5, element.polarity())
        except AttributeError:
            pass
        else:
            patch_start_y = x

        rectx = patches.Rectangle(
            (patch_start_x, patch_start_y),
            patch_width,
            patch_height,
            linewidth=0.1,
            # label=label,
            edgecolor='white',
            facecolor=colour,
            alpha=alpha,
        )
        xcentre = patch_start_x + 0.5 * patch_width
        ycentre = patch_start_y + 0.5 * patch_height
        t2 = mpl.transforms.Affine2D().rotate_around(xcentre, ycentre, -row.theta) + axes.transData
        rectx.set_transform(t2)
        axes.add_patch(rectx)

        if annotate:
            element_type = row.keyword

            annotation = axes.annotate(
                f"{element_type}: {row.name}",
                xy=(z, x),  # xycoords='data',
                xytext=(z, x),
                textcoords='data',
                horizontalalignment="left",
                arrowprops=dict(arrowstyle="simple", connectionstyle="arc3,rad=+0.2"),
                bbox=dict(boxstyle="round", facecolor="w", edgecolor="0.5", alpha=0.9),
                fontsize=18,
            )
            # by default, disable the annotation visibility
            annotation.set_visible(False)
            patch_ids_to_annotations[id(rectx)] = annotation
            patch_ids_to_elements[id(rectx)] = row
            rectx.set_picker(True)

    def on_pick(event):
        annotation = patch_ids_to_annotations[id(event.artist)]
        annotation.set_visible(not annotation.get_visible())
        # adjust_text([a for a in patch_ids_to_annotations.values()
        #              if a.get_visible()])
        fig.canvas.draw_idle()  # text_from_points=False
        element = patch_ids_to_elements[id(event.artist)]
        print(repr(element))

        # from IPython import embed; embed()

        # plt.draw()

    def on_press(event):
        if event.key == 'c':  # press c to clear the annotations.
            for annotation in patch_ids_to_annotations.values():
                annotation.set_visible(False)
        fig.canvas.draw_idle()

    fig.canvas.mpl_connect('pick_event', on_pick)
    fig.canvas.mpl_connect('key_press_event', on_press)


def draw(
    fig,
    axes,
    sequence,
    colour_map=None,
    annotate=True,
    magnet_width=MAGNET_WIDTH,
    coords="curvilinear",
    s_offset=0,
    **drawlinekw,
):
    draw_line(axes, sequence, s_offset=s_offset, **drawlinekw)
    if colour_map is None:
        colour_map = DEFAULT_COLOUR_MAP

    patch_ids_to_annotations = {}
    patch_ids_to_elements = {}
    # from IPython import embed; embed()
    
    for element, surv in zip(sequence, sequence.survey().itertuples()):
        colour = DEFAULT_COLOUR_MAP[surv.keyword]
        alpha = 1

        if not element.is_active():
            alpha = 0.25

        if colour is None:
            alpha = 0.0

        # This is the end basically.
        length = element.length

        if coords == "curvilinear":
            element_x1 = surv.s + s_offset
            element_x0 = element_x1 - length
        elif coords == "cartesian":
            element_x1 = surv.z + s_offset
            element_x0 = element_x1 - surv.zlocal[2] * length
        

        length = element.length
        patch_start_x = element_x0
        patch_width = element_x1 - element_x0
        patch_start_y = - 2 * magnet_width
        patch_height = 4 * magnet_width

        try:
            patch_height = copysign(patch_height * 0.5, element.polarity())
        except AttributeError:
            pass
        else:
            patch_start_y = 0 # patch_start_x

        rectx = patches.Rectangle(
            (patch_start_x, patch_start_y),
            patch_width,
            patch_height,
            linewidth=0.1,
            edgecolor='white',
            facecolor=colour,
            alpha=alpha,
        )
        axes.add_patch(rectx)

        if annotate:
            element_type = type(element).__name__

            annotation = axes.annotate(
                f"{element_type}: {element.name}",
                xy=(patch_start_x, 0),  # xycoords='data',
                xytext=(patch_start_x, 0),
                textcoords='data',
                horizontalalignment="left",
                arrowprops=dict(arrowstyle="simple", connectionstyle="arc3,rad=+0.2"),
                bbox=dict(boxstyle="round", facecolor="w", edgecolor="0.5", alpha=0.9),
                fontsize=18,
            )
            # by default, disable the annotation visibility
            annotation.set_visible(False)
            patch_ids_to_annotations[id(rectx)] = annotation
            patch_ids_to_elements[id(rectx)] = element
            rectx.set_picker(True)

    def on_pick(event):
        annotation = patch_ids_to_annotations[id(event.artist)]
        annotation.set_visible(not annotation.get_visible())
        # adjust_text([a for a in patch_ids_to_annotations.values()
        #              if a.get_visible()])
        fig.canvas.draw_idle()  # text_from_points=False
        element = patch_ids_to_elements[id(event.artist)]

        print(repr(element))

        # plt.draw()

    def on_press(event):
        if event.key == 'c':  # press c to clear the annotations.
            for annotation in patch_ids_to_annotations.values():
                annotation.set_visible(False)
        fig.canvas.draw_idle()

    fig.canvas.mpl_connect('pick_event', on_pick)
    fig.canvas.mpl_connect('key_press_event', on_press)


def draw_line(axes, sequence, s_offset=0, **plotkw):
    survey = sequence.survey()

    s = np.concatenate(([0], survey.s))
    s += s_offset

    x = np.zeros_like(s)

    axes.plot(s, x, **plotkw)
    axes.set_ylim(-3, 3)


def draw_survey_line(axes, z, rcoord, **plotkw):
    axes.plot(z, rcoord, **plotkw)



import pandas as pd
import numpy as np
from matplotlib.patches import Patch
from collections.abc import Sequence

import logging

logger = logging.getLogger(__name__)

MAGNET_WIDTH = 0.1

# DEFAULT_COLOUR_MAP = Quadrupole

# If this was actually good: the survey would be generated AUTOMATICALLY from
# the beam angles and tilts! And some offset (with angle) would just belong to
# the beamline! instance. And could maybe even support different visualistion
# backends 🤔.

# Actually to get rotations to work properly, this is more or less necessary,
# because otherwise I have no way of knowing the orientation of a given element
# in XYZ space (survey) without inspecting surrounding elements. So could also
# just be an interface change, without so much work, but i think a
# "beamline.getRotation(element)" etc approach would be the best approach.


class Beamline(Sequence):
    def __init__(self, items):
        self._sequence = list(items)

    def __getitem__(self, key):
        return self._sequence[key]

    def __len__(self):
        return len(self._sequence)

    def survey(self):
        initial_offset = np.array([0, 0, 0]).reshape(3, 1)
        initial_rotation = np.identity(3)

        global_rotations = [initial_rotation]
        global_placements = [initial_offset]

        for element in self:
            local_displacement = element.displacement()
            local_rotation = element.rotation()

            # W_{i-1} in MADX manual
            previous_rotation = global_rotations[-1]
            previous_placement = global_placements[-1]

            # V_i in MADX manual
            global_placement = previous_rotation @ local_displacement + previous_placement
            global_rotation = previous_rotation @ local_rotation

            global_placements.append(global_placement)
            global_rotations.append(global_rotation)

        global_placements = np.hstack(global_placements)
        x, y, z = global_placements[..., 1:]

        active = [x.is_active() for x in self]
        length = [x.length for x in self]

        theta, _, _ = rotation_matrices_to_madx_rotations(global_rotations[1:])
        xlocal, ylocal, zlocal = rotation_matrices_to_local_axis(global_rotations[1:])

        df = pd.DataFrame.from_dict(
            {
                "name": self.name(),
                "keyword": self.keyword(),
                "x": x,
                "y": y,
                "z": z,
                "xlocal": xlocal,
                "ylocal": ylocal,
                "zlocal": zlocal,
                "s": self.s(),
                "theta": theta,
                "length": length,
                "active": active,
                "e1": self._get_attr_or_zero("e1"),
                "e2": self._get_attr_or_zero("e2"),
                "k1": self._get_attr_or_zero("k1"),
                "angle": self._get_attr_or_zero("angle"),
            }
        )

        return df

    def _get_attr_or_zero(self, attr_name):
        result = []
        for x in self:
            try:
                result.append(getattr(x, attr_name))
            except AttributeError:
                result.append(0.0)
        return result

    def keyword(self):
        return [type(element).__name__ for element in self]

    def name(self):
        return [element.name for element in self]

    def s(self):
        return np.cumsum([element.length for element in self])


def rotation_matrices_to_madx_rotations(matrices):
    thetas = []
    unit_z = [0, 0, 1]
    # import ipdb; ipdb.set_trace()
    for matrix in matrices:
        # New direction, starting from pointing along global z:
        # vec = matrix @ unit_z
        new_z = matrix[:, 2]
        # Project the new z onto the x-z plane
        new_z[1] = 0.0

        theta = np.arccos(np.dot(unit_z, new_z) / np.linalg.norm(new_z))
        thetas.append(theta)

    return thetas, [], []


def rotation_matrices_to_local_axis(matrices):
    x = []
    y = []
    z = []
    for matrix in matrices:
        # Third column is new z axis (i.e. comoving s)
        new_x = matrix[:, 0]
        new_y = matrix[:, 1]
        new_z = matrix[:, 2]

        x.append(new_x)
        y.append(new_y)
        z.append(new_z)

    return x, y, z


class StraightElementMixIn:
    def displacement(self):
        return np.array([0, 0, self.length]).reshape(3, 1)

    def rotation(self):
        return np.identity(3)


class Element:
    def __init__(self, name, length, **misc):
        """For now position is the END?  survey and S can be dengerate, x and y
        just stay 0...

        """
        self.name = name
        # self.position = np.array(position)
        self.length = length
        self.misc = misc
        # self.tilt = 0

    def __repr__(self):
        typename = type(self).__name__
        rstr = f"<{typename}: name={self.name}, l={self.length}"
        if self.misc:
            rstr += ", " + repr(self.misc)
        rstr += ">"
        return rstr

    def __str__(self):
        return f"{type(self).__name__}: {self.name}"

    def is_active(self):
        return True

    # def tilt_matrix()


class ThinElement(Element, StraightElementMixIn):
    def __init__(self, name, **misc):
        super().__init__(name, 0.0, **misc)

    def is_active(self):
        return True

    def __repr__(self):
        typename = type(self).__name__
        return f"<{typename}: name={self.name}>"


class Marker(ThinElement):
    pass


class Monitor(ThinElement):
    pass


class Drift(Element, StraightElementMixIn):
    pass


class SimpleDipole(Element):
    def __init__(self, name, length, angle, **misc):
        super().__init__(name, length, **misc)
        self.angle = angle

    def is_active(self):
        return self.angle != 0

    def displacement(self):
        if self.angle == 0:
            res = np.array([0, 0, self.length])
        else:
            res = np.array([self.rho * (np.cos(self.angle) - 1), 0, self.rho * np.sin(self.angle)])

        return res.reshape(3, 1)

    def rotation(self):
        return np.array(
            [[np.cos(self.angle), 0, -np.sin(self.angle)], [0, 1, 0], [np.sin(self.angle), 0, np.cos(self.angle)]]
        )

    def __repr__(self):
        typename = type(self).__name__
        return f"<{typename}: name={self.name}, l={self.length}, angle={self.angle}>"


class RBend(SimpleDipole):
    @property
    def rho(self):
        "Using the MAD8/X definition here..."
        return self.length / (2 * np.sin(self.angle / 2))

    def is_square(self):
        return (self.e1 == 0) and (self.e2 == 0)


class SBend(SimpleDipole):
    @property
    def rho(self):
        "Using the MAD8/X definition here..."
        return self.length / self.angle

    def is_square(self):
        return (self.e1 == self.angle / 2) and (self.e2 == self.angle / 2)


class HKicker(SimpleDipole, StraightElementMixIn):
    pass


class VKicker(SimpleDipole, StraightElementMixIn):
    pass


class Kicker(SimpleDipole, StraightElementMixIn):
    pass


class Quadrupole(Element, StraightElementMixIn):
    def __init__(self, name, length, k1, **misc):
        super().__init__(name, length, **misc)
        self.k1 = k1

    def is_active(self):
        return self.k1 != 0

    def __str__(self):
        l = self.length
        k1 = self.k1
        k1l = l * k1
        return f"<Quad: {self.name}, {l=}, {k1=}, {k1l=}>"

    def polarity(self):
        return np.sign(self.k1)


class Sextupole(Element, StraightElementMixIn):
    def __init__(self, name, length, k2, **misc):
        super().__init__(name, length, **misc)
        self.k2 = k2

    def is_active(self):
        return self.k2 != 0


class Octupole(Element, StraightElementMixIn):
    def __init__(self, name, length, k3, **misc):
        super().__init__(name, length, **misc)
        self.k3 = k3

    def is_active(self):
        return self.k3 != 0


class RFCavity(Element, StraightElementMixIn):
    def __init__(self, name, length, voltage, phase, **misc):
        super().__init__(name, length, **misc)
        self.voltage = voltage
        self.phase = phase

    def is_active(self):
        return self.voltage != 0


class Solenoid(Element, StraightElementMixIn):
    def __init__(self, name, length, ks, **misc):
        super().__init__(name, length, **misc)
        self.ks = ks

    def is_active(self):
        return self.ks != 0


class Collimator(Element, StraightElementMixIn):
    pass


class Cavity(Element):
    pass


class GenericMap(Element):
    pass


class TransverseDeflectingCavity(Element, StraightElementMixIn):
    def __init__(self, name, length, voltage, **misc):
        super().__init__(name, length, **misc)
        self.voltage = voltage

    def is_active(self):
        return self.voltage != 0


class Undulator(Element, StraightElementMixIn):
    def __init__(self, name, length, kx=0.0, ky=0.0, **misc):
        super().__init__(name, length, **misc)
        self.kx = kx
        self.ky =ky

    def is_active(self):
        return self.kx !=0 or self.ky !=0
        
        


